(***************************************************************************
  Serial
  Serial port stream
  PJB October 20, 1993, CompuServe mail to INTERNET:d91-pbr@nada.kth.se
  Copyright PJB 1993, All Rights Reserved.
  Free source, use at your own risk.
  If modified, please state so if you pass this around.

  No flow control, only one COM port at a time
  The BufSize values must be one less than a power of two,
  eg 127, 255, 511 etc.
  The usual IRQ problem, COM1 and 3 conflict, and 2 and 4

  The assembler interrupt handler manages at 115200 bps on a 286/6

***************************************************************************)
unit Serial;
{$IFDEF VER60}
 {$B-,X+}
{$ELSE}
 {$B-,Q-,X+}
{$ENDIF}

interface

  uses
    Objects;

  const
    InBufSize  = 255;
    OutBufSize = 127;

    (*******************************************************************
      Modem status (MSR)
    *******************************************************************)
    DeltaCTS     = $01;
    DeltaDSR     = $02;
    DeltaRing    = $04;
    DeltaCarrier = $08;
    CTS          = $10;
    DSR          = $20;
    Ring         = $40;
    Carrier      = $80;

    (*******************************************************************
      Modem control (MCR)
    *******************************************************************)
    DTR      = $01;
    RTS      = $02;
    OUT2     = $08;
    LoopBack = $10;

    (*******************************************************************
      Interrupt Enable (IER)
    *******************************************************************)
    ieReceive       = $01;
    ieTransmit      = $02;
    ieReceiveError  = $04;
    ieModemStatus   = $08;

    (*******************************************************************
      Line Status Register (LSR)
    *******************************************************************)
    lsOverrunError  = $02;
    lsParityError   = $04;
    lsFramingError  = $08;
    lsBreakDetect   = $10;
    lsTxEmpty       = $20;
    lsAllTxEmpty    = $40;

    (*******************************************************************
      Predefined IRQ values for COM1-COM4
    *******************************************************************)
    IRQList : array [1..4] of Word = (4,3,4,3);


  type
    ParityType = (None, Odd, __z1, Even, __z2, Mark, __z3, Space);

    PByte  = ^Byte;

    COMBufferIn  = array [0..InBufSize]  of Byte;
    COMBufferOut = array [0..OutBufSize] of Byte;

    PSerialStream = ^TSerialStream;
    TSerialStream =
      object (TStream)
        IOPort        : Word;
        IRQNumber     : Word;
        Number        : integer;

        Baud          : longint;
        Parity        : ParityType;
        StopBits      : integer;
        DataBits      : integer;

        OverrunErrors : Word;
        ParityErrors  : Word;
        FramingErrors : Word;
        BreakErrors   : Word;
        ErrorCount    : Word;

        Using16550    : Boolean;
        ModemStatus   : Byte;

        InHead        : Word;
        InTail        : Word;
        OutHead       : Word;
        OutTail       : Word;

        InBuffer      : COMBufferIn;
        OutBuffer     : COMBufferOut;

        constructor Init(aNumber:Integer);
        constructor InitPort(Port, IRQ:Word);
        destructor Done; virtual;

        function  BytesInInputBuffer:Word;
        function  BytesInOutputBuffer:Word;
        procedure ClearErrors;
        procedure ClearInBuffer;
        procedure ClearOutBuffer;
        function  DataReady:Boolean;
        procedure EnableInterrupts(Ints:Byte);
        procedure GetBaud(var aBaud:longint; var aParity:ParityType; var aDataBits, aStopBits:integer);
        procedure HandleInterrupt;
        procedure InstallInt;
        function  GetModemStatus:Byte;
        function  LoopBackTest:Boolean;
        procedure ModifyReg(Reg:Word; Value:Byte; Enable:Boolean);
        procedure Read(var Buf; Count:Word); virtual;
        procedure RemoveInt;
        procedure SetBaud(aBaud:Longint; aParity:ParityType; aDataBits, aStopBits:Integer);
        procedure Use16550(Enable:Boolean);
        procedure Write(var Buf; Count:Word); virtual;
        procedure WriteString(const S:String); virtual;
        procedure WriteNow(var Buf; Count:Word);
      end;


(***************************************************************************
***************************************************************************)
implementation


  uses
    Dos;

  const
   {$IFDEF VER60}
    Seg0040 = $0040;
   {$ENDIF}

    (*******************************************************************
      IOPort offset constants
    *******************************************************************)
    THR = 0;    { Transmit Holding Register  }
    RHR = 0;    { Receive Holding Register   }
    DLL = 0;    { Divisor Latch Register LSB }
    DLM = 1;    { Divisor Latch Register MSB }
    IER = 1;    { Interrupt Enable Register  }
    IIR = 2;    { Interrupt ID Register      }
    FCR = 2;    { 16550 FIFO Control Register}
    LCR = 3;    { Line Control Register      }
    MCR = 4;    { Modem Control Register     }
    LSR = 5;    { Line Status Register       }
    MSR = 6;    { Modem Status Register      }

    (*******************************************************************
      Interrupt Identification (IIR)
    *******************************************************************)
    iiModemStatus      = $00;
    iiTransmitRegEmpty = $02;
    iiReceiveChar      = $04;
    iiReceiveError     = $06;

    (*******************************************************************
      Line Control (LCR)
    *******************************************************************)
    lcDataBits   = $03;
    lcStopBits   = $04;
    lcParityBits = $38;
    lcSetBreak   = $40;
    lcDLAB       = $80;


    (*******************************************************************
      PC hardware interrupt controller
    *******************************************************************)
    PICMask        = $21;   { Programmable interrupt controller IRQ mask }
    PICCommand     = $20;   { PIC command address }
    EndOfInterrupt = $20;   { EndOfInterrupt signal }


  (*******************************************************************
    State tracking variables
  *******************************************************************)
  var
    COMPort        : PSerialStream;
    OldIntVector   : Pointer;
    IRQMask        : Byte;
    OldIRQMask     : Byte;
    SaveExit       : Pointer;


  (*******************************************************************
    Inline code
  *******************************************************************)
  procedure EnableInts;  INLINE($FB);
  procedure DisableInts; INLINE($FA);


(***************************************************************************
  Serial Port object
***************************************************************************)

  constructor TSerialStream.Init;
  begin
    Number:=aNumber;
    InitPort(MemW[Seg0040:(Number-1) SHL 1], IRQList[Number]);
  end;

  constructor TSerialStream.InitPort;
  begin
    IOPort:=Port;
    IRQNumber:=IRQ;
    InstallInt;
  end;

  destructor TSerialStream.Done;
  begin
    RemoveInt;
    if Using16550 then
      Use16550(False);
  end;

  function TSerialStream.BytesInInputBuffer:Word;
  begin
    BytesInInputBuffer:=(InTail-InHead) and InBufSize;
  end;

  function TSerialStream.BytesInOutputBuffer:Word;
  begin
    BytesInOutputBuffer:=(OutTail-OutHead) and InBufSize;
  end;

  procedure TSerialStream.ClearErrors;
  begin
    OverrunErrors:=0;
    ParityErrors:=0;
    FramingErrors:=0;
    BreakErrors:=0;
    ErrorCount:=0;
  end;

  procedure TSerialStream.ClearInBuffer;
  begin
    DisableInts;
    InHead:=0;
    InTail:=0;
    EnableInts;
  end;

  procedure TSerialStream.ClearOutBuffer;
  begin
    DisableInts;
    OutHead:=0;
    OutTail:=0;
    EnableInts;
  end;


  (*******************************************************************
    Is data available in incoming buffer
  *******************************************************************)
  function TSerialStream.DataReady;
  begin
    DataReady:=InTail<>InHead;
  end;


  (*******************************************************************
    Enable serial interrupt types like line status and receiver error
  *******************************************************************)
  procedure TSerialStream.EnableInterrupts;
  begin
    ModifyReg(IER, Ints, True);
  end;

  (*******************************************************************
    Get current COM port setting directly from port
  *******************************************************************)
  procedure TSerialStream.GetBaud;
    var
      temp : Byte;
  begin
    DisableInts;
    ModifyReg(LCR, lcDLAB, True);

    aBaud:=115200 div PortW[IOPort+DLL];

    temp:=Port[IOPort+LCR];
    aDataBits:=(temp and lcDataBits)+5;
    aStopBits:=((temp and lcStopBits) shr 2)+1;
    aParity:=ParityType((temp and lcParityBits) shr 3);

    ModifyReg(LCR, lcDLAB, False);
    EnableInts;
  end;


  (*******************************************************************
    Get line status (Ring, Carrier etc)
  *******************************************************************)
  function TSerialStream.GetModemStatus;
  begin
    GetModemStatus:=Port[IOPort+MSR];
  end;


  (*******************************************************************
    Perform Loopback test to see if port performs alright
  *******************************************************************)
  function TSerialStream.LoopBackTest;
   {$IFOPT Q+} {$DEFINE QDefined} {$ENDIF} {$Q-}
    function WaitForResponse:Boolean;
      var
        Fin : Byte;
    begin
      Fin:=Mem[Seg0040:$6C]+2;
      while (Fin-Mem[Seg0040:$6C]>0) and (Port[IOPort+IIR] and $07<>iiReceiveChar) do
        ;
      WaitForResponse:=Port[IOPort+IIR] and $07=iiReceiveChar;
    end;
   {$IFDEF QDefined} {$Q+} {$ENDIF}

    const
      SS = 'Hello!';
      S : string [Length(SS)] = SS;
    var
      i : integer;
      Loop : Boolean;
  begin
    ModifyReg(MCR, LoopBack, True);

    i:=1;
    Port[IOPort+THR]:=Ord(S[I]);

    LoopBackTest:=False;
    Loop:=True;

    while Loop and WaitForResponse do
    begin
      if S[i]<>Chr(Port[IOPort+RHR]) then
        Loop:=False
      else
      begin
        Inc(i);
        if i>Length(S) then
        begin
          LoopBackTest:=True;
          Loop:=False;
        end
        else
          Port[IOPort+THR]:=Ord(S[I]);
      end;
    end;

    ModifyReg(MCR, LoopBack, False);
  end;


  (*******************************************************************
    Modify a UART register
  *******************************************************************)
  procedure TSerialStream.ModifyReg(Reg:Word; Value:Byte; Enable:Boolean);
    var
      x : Byte;
  begin
    x:=Port[IOPort+Reg];
    if Enable then
      x:=x or Value
    else
      x:=x and not Value;
    Port[IOPort+Reg]:=x;
  end;


  (*******************************************************************
    Read a block of characters
  *******************************************************************)
  procedure TSerialStream.Read(var Buf; Count:Word);
    var
      Dest : PByte;
  begin
    Dest:=PByte(@Buf);

    while Count>0 do
    begin
      repeat until InTail<>InHead;

      Dest^:=InBuffer[InHead];
      InHead:=(InHead+1) and InBufSize;
      Inc(Dest);
      Dec(Count);
    end;
  end;


  (*******************************************************************
    Set baud rate, parity, data and stopbits
  *******************************************************************)
  procedure TSerialStream.SetBaud;
    var
      Divisor : Word;
  begin
    Baud:=aBaud;
    Divisor:=115200 div Baud;
    Parity:=aParity;
    DataBits:=aDataBits;
    StopBits:=aStopBits;

    DisableInts;
    Port[IOPort+LCR]:=lcDLAB+
                      (DataBits-5) and lcDataBits or
                      (((StopBits-1) shl 2) and lcStopBits) or
                      ((Ord(Parity) shl 3) and lcParityBits);
    Port[IOPort+DLL]:=Lo(Divisor);
    Port[IOPort+DLM]:=Hi(Divisor);
    ModifyReg(LCR, lcDLAB, False);
    EnableInts;
  end;


  (*******************************************************************
    UART 16550 FIFO queue on/off
  *******************************************************************)
  procedure TSerialStream.Use16550;
  begin
    if Enable then
      Port[IOPort+FCR]:=$C7
    else
      Port[IOPort+FCR]:=0;

    Using16550:=Port[IOPort+IIR] and $40 <> 0;
  end;


  (*******************************************************************
    Write a block of bytes, buffered, using interrupts
  *******************************************************************)
  procedure TSerialStream.Write;
    var
      B      : Word;
      Source : PByte;
  begin
    Source:=PByte(@Buf);
    while Count>0 do
    begin
      B:=OutBufSize-OutTail+1;
      if Count<B then
        B:=Count;
      Move(Source^, OutBuffer[OutTail], B);
      OutTail:=(OutTail+B) and OutBufSize;

      ModifyReg(IER, ieTransmit, True);
      Dec(Count, B);
      Inc(Source, B);
    end;
  end;


  (*******************************************************************
    Write a string
  *******************************************************************)
  procedure TSerialStream.WriteString;
  begin
    Write(PString(@S)^[1], Length(S));
  end;


  (*******************************************************************
    Write block directly to port, no interrupts
  *******************************************************************)
  procedure TSerialStream.WriteNow;
    var
      Time : Word;
      TimeOut, Ready : Boolean;
      Source : PByte;
  begin
    TimeOut:=False;
    Source:=PByte(@Buf);

    while (Count>0) and not TimeOut do
    begin
      Time:=MemW[Seg0040:$6C]+2;
      repeat
        Ready:=(Port[IOPort+LSR] and $20) <> 0;
      until Ready or (MemW[Seg0040:$6C]=Time);

      if Ready then
        Port[IOPort+THR]:=Source^
      else
        TimeOut:=True;

      Dec(Count);
      Inc(Source);
    end;
  end;


  (*******************************************************************
    High level interrupt handler
    Too slow for 115200 bps on a 386DX20
  *******************************************************************)
  procedure TSerialStream.HandleInterrupt;
    var
      Temp : Byte;
  begin
    Temp:=Port[COMPort^.IOPort+IIR];
    if Temp and 1=0 then
      case Temp and $06 of
        iiReceiveChar:
          if (InTail+1) and InBufSize<>InHead then
          begin
            InBuffer[InTail]:=Port[IOPort+RHR];
            InTail:=(InTail+1) and InBufSize;
          end;

        iiTransmitRegEmpty:
          if OutHead<>OutTail then
          begin
            Port[IOPort+THR]:=OutBuffer[OutHead];
            OutHead:=(OutHead+1) and OutBufSize;
          end
          else
            Port[IOPort+IER]:=Port[IOPort+IER] and not ieTransmit;

        iiModemStatus:  ModemStatus:=Port[IOPort+MSR];

        iiReceiveError:
          begin
            Temp:=Port[IOPort+LSR];
            if (Temp and lsOverrunError<>0) then Inc(OverrunErrors);
            if (Temp and lsParityError<>0)  then Inc(ParityErrors);
            if (Temp and lsFramingError<>0) then Inc(FramingErrors);
            if (Temp and lsBreakDetect<>0)  then Inc(BreakErrors);
            Inc(ErrorCount);
          end;
      end
  end;


  (*******************************************************************
    Interface to the high level interrupt handler
  *******************************************************************)
  procedure IntHandler2; interrupt;
  begin
    Port[PICCommand]:=$0B;
    if Port[PICCommand] and IRQMask<>0 then
    begin
      ComPort^.HandleInterrupt;
      Port[PICCommand]:=EndOfInterrupt;
    end
    else
    asm
      pushf
      call  OldIntVector
    end;
  end;


  (*******************************************************************
    Efficient assembler interrupt handler
  *******************************************************************)
  procedure IntHandler; assembler;
  asm
      push ax
      push bx
      push dx
      push di
      push ds

      mov  ax,SEG @Data
      mov  ds,ax

      mov  al,0Bh
      out  PICCommand,al

      in   al,PICCommand
      and  al,IRQMask
      je   @ChainInt

      lds  di,COMPort

    @CheckInt:
      mov  dx,[di].TSerialStream.IOPort
      add  dl,IIR
      in   al,dx

      test al,1
      jne  @Fin

      and  ax,06
      mov  bx,ax
      jmp  WORD PTR cs:[bx+OFFSET @JumpTable]

    @JumpTable:
      dw   OFFSET @ModemStatus
      dw   OFFSET @WriteChar
      dw   OFFSET @ReadChar
      dw   OFFSET @Error

    @ChainInt:
      pushf
      call OldIntVector
      jmp  @Exit

    @WriteChar:
      mov  bx,[di].TSerialStream.OutHead
      cmp  bx,[di].TSerialStream.OutTail
      je   @Fin

      add  dl,THR-IIR
      mov  al,[di+bx].TSerialStream.OutBuffer.Byte
      out  dx,al

      inc  bx
      and  bx,OutBufSize
      mov  [di].TSerialStream.OutHead,bx
      jmp  @CheckInt

    @ModemStatus:
      add  dl,MSR-IIR
      in   al,dx
      mov  [di].TSerialStream.ModemStatus,al
      jmp  @CheckInt

    @Error:
      add  dl,LSR-IIR
      in   al,dx

      mov  bx,0
      shr  al,1
      shr  al,1
      adc  [di].TSerialStream.OverrunErrors,bx
      shr  al,1
      adc  [di].TSerialStream.ParityErrors,bx
      shr  al,1
      adc  [di].TSerialStream.FramingErrors,bx
      shr  al,1
      adc  [di].TSerialStream.BreakErrors,bx

      inc  [di].TSerialStream.ErrorCount
      jmp  @CheckInt

    @ReadChar:
      add  dl,RHR-IIR
      mov  bx,[di].TSerialStream.InTail

      mov  ax,bx
      inc  ax
      and  ax,InBufSize
      cmp  ax,[di].TSerialStream.InHead
      je   @Fin                                (* Buffer full *)

      in   al,dx
      mov  [di+bx].TSerialStream.InBuffer.Byte,al

      inc  bx
      and  bx,InBufSize
      mov  [di].TSerialStream.InTail,bx
      jmp  @CheckInt

    @Fin:
      mov  al,EndOfInterrupt
      out  PICCommand,al

    @Exit:
      pop  ds
      pop  di
      pop  dx
      pop  bx
      pop  ax
      iret
  end;


  (*******************************************************************
    Exit handler uninstalls interrupt hook
  *******************************************************************)
  procedure ErrorExit; far;
  begin
    ExitProc:=SaveExit;
    if COMPort<>Nil then
      COMPort^.Done;
  end;


  (*******************************************************************
    Hook serial interrupt
  *******************************************************************)
  procedure TSerialStream.InstallInt;
    var
      Temp : Byte;
  begin
    DisableInts;

   {$IFDEF VER60}
    IRQMask:=0;
   {$ENDIF}

    COMPort:=@Self;

    (* Reset pending ints *)
    Temp:=Port[IOPort+LSR];
    Temp:=Port[IOPort+RHR];

    (* Check if int connected *)
    Temp:=1 SHL IRQNumber;

    if (IRQMask and Temp)=0 then
    begin
      IRQMask:=IRQMask or Temp;
      GetIntVec($08+IRQNumber, OldIntVector);
      SetIntVec($08+IRQNumber, @IntHandler);

      (* Enable 8259 IRQ handling *)
      OldIRQMask:=Port[PICMask];
      Port[PICMask]:=OldIRQMask and not Temp;
    end;

    { Enable serial interrupts }
    Port[IOPort+IER]:=ieReceive;      { Receive interrupt on }
    Port[IOPort+MCR]:=DTR+RTS+OUT2;   { DTR, RTS and OUT2 on }
    EnableInts;

    SaveExit:=ExitProc;
    ExitProc:=@ErrorExit;
  end;


  (*******************************************************************
    Remove interrupt hook
  *******************************************************************)
  procedure TSerialStream.RemoveInt;
    var
      Temp : Byte;
  begin
    DisableInts;

    { Disable serial interrupts }
    Port[IOPort+IER]:=0;   { No interrupt types }
    Port[IOPort+MCR]:=0;   { DTR, RTS, OUT2 off }

    Temp:=1 SHL IRQNumber;
    if (IRQMask and Temp)<>0 then
    begin
      Port[PICMask]:=OldIRQMask;
      SetIntVec($08+IRQNumber, OldIntVector);
      IRQMask:=IRQMask and not Temp;
    end;
    EnableInts;
  end;


(***************************************************************************
***************************************************************************)
end.


